/*********************************************************************
 *** halockrun.c
 *** $Id: halockrun.c,v 1.26 2007/08/18 15:12:59 mws Exp $
 ********************************************************************/
/*
HA-Tools (c) 2001,2003,2005-2007 by Markus Winand <mws@fatalmind.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include "halockrun.h"
#include "error.h"

extern char* optarg;
extern int optind;

int FAIL_EXITCODE = _FAIL_EXITCODE;
int FAIL_EXITCODE_NB = _FAIL_EXITCODE;
char *cmdname = NULL;

/*
 *
 * fork and exec
 *
 */

int mk_child(int fork_mode, char *argv[])
{
	int childstatus = 0;
	pid_t childpid = 0;
	struct sigaction act;
	int ret;

	if (fork_mode) {
		childpid = fork();
	}
	switch (childpid) {
	case -1: /* fail */
		error("fork() failed: %s\n", strerror(errno));
		break;
	case 0: /* fork() succeeded */
		/* kiddie */

		execvp(argv[0], argv);

		/* exec failed */
		error("exec(%s) failed: %s\n", argv[0], strerror(errno));
		/* never returns */
		break;
	default: /* parent */
		/*
		 *
		 * Set the signalhandlers for parent to ignore signals 
		 * in fork_mode
		 *
		 */
		act.sa_handler = SIG_IGN;
		sigemptyset(&act.sa_mask);
		act.sa_flags = 0;

		sigaction(SIGTERM, &act, NULL);
		sigaction(SIGHUP,  &act, NULL);
		sigaction(SIGINT,  &act, NULL);
		sigaction(SIGPIPE, &act, NULL);
		sigaction(SIGQUIT, &act, NULL);
		sigaction(SIGUSR1, &act, NULL);
		sigaction(SIGUSR2, &act, NULL);

		do {
			ret = waitpid(childpid, &childstatus, 0);
		} while (ret == -1 && errno == EINTR);

		break;
	};
	if (WIFEXITED(childstatus)) {
		return WEXITSTATUS(childstatus);
	} else {
		return FAIL_EXITCODE;
	}
}


/*
 *
 * print usage and exit
 *
 */

void usage() {
	fprintf(stderr, "usage: %s [-ancf] [-e nbexitcode] [-E exitcode]"
			"lockfile command [args]\n", cmdname);
	fprintf(stderr, "       %s -t [-c] [-e exitcode] "
			"lockfile\n", cmdname);
	exit(FAIL_EXITCODE);
}

void longusage() {
	fprintf(stderr, 
"usage: %s [-ancf] [-e exitcode] [-E nexitcode] lockfile command [args]\n"
"       %s -t [-c] [-e exitcode] [-E nexitcode] lockfile\n"
"       %s [-h|-?]\n"
"Options:\n",
cmdname, cmdname, cmdname);
	fprintf(stderr,
"  -a            Async mode. Starts %s in the background\n"
"  -e exitcode   Changes the exitcode returned by %s if an error occured.\n"
"  -E nexitcode  Changes the exitcode returned by %s in case the lock could not\n"
"                obtained. Defaults to the values of -e of omitted.\n", cmdname, cmdname, cmdname);

	fprintf(stderr, 
"  -n            Nonblocking. Exits immediately if the lock cannot be obtained\n"
"  -c            Creates the lockfile if it doesn't exist\n"
"  -f            Fork mode. Runs %s and the command in differnt processes\n"
"  -t            Test mode. %s will check the lockfile and report the process\n"
"                who holds the lock if it's locked\n\n",
cmdname, cmdname);
        fprintf(stderr,   
"Version:\n"
"  V%s\n"
"  Copyright (c) 2001,2003,2005-2007 by Markus Winand <mws@fatalmind.com>\n"
"  This program is distributed in the hope that it will be useful,\n"
"  but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
"  GNU General Public License for more details.\n\n", VERSION);

	exit(FAIL_EXITCODE);
}

/*
 * global config parameters
 */

int async = 0;
int fork_mode = 0;
int test_mode = 0;
int openmode = O_WRONLY;
int lockmode = F_SETLKW;
char *lockfile = NULL;

/*
 * 
 * command line parsing
 *
 */

char **cmd_line(int argc, char *argv[]) {
	char **childargv = NULL;
	int childargc = argc-1;
	int hlp;
	int ch;
	char dummy;

	/* was argument -e or -E specified on the cmd line? */
	int argE = 0;
	int arge = 0;

#ifdef HAVE_SETENV
	setenv("POSIXLY_CORRECT", "1", 0);
#endif

	while ((ch = getopt(argc,argv, "ance:E:h?ft")) != -1) {
		switch (ch) {
		case 'a':
			async = 1;
			break;
		case 'n':
			lockmode = F_SETLK;
			break;
		case 't':
			test_mode = 1;	
			break;
		case 'c':
			openmode |= O_CREAT;
			break;
		case 'e':
			if (sscanf(optarg, "%d%c", &hlp, &dummy) != 1) {
				error("illegal option \"-e %s\"\n",
						optarg);
			}
			FAIL_EXITCODE = hlp;
			arge = 1;
			break;
		case 'E':
			if (sscanf(optarg, "%d%c", &hlp, &dummy) != 1) {
				error("illegal option \"-E %s\"\n",
						optarg);
			}
			FAIL_EXITCODE_NB = hlp;
			argE = 1;
			break;
		case 'h':
		case '?':
			longusage();
			break;
		case 'f':
			fork_mode = 1;
			break;
		default:
			usage();
			exit(FAIL_EXITCODE);
		};
	}
 		
	childargc = argc - optind;
	childargv = argv + optind;

	if ((! test_mode && childargc < 2) || (test_mode && (childargc != 1))) {
		usage();
	}

	/* if -e was specified but not -E */
	if (arge && ! argE) {
		/* we use the -e setting also for -E (compatibility) */
		FAIL_EXITCODE_NB = FAIL_EXITCODE;
	}
	if (test_mode && ! argE) {
		/* the default nexitcode in test mode */
		FAIL_EXITCODE_NB = 1;
	}
	lockfile = childargv[0];
	childargv++;
	childargc--;

	childargv[childargc] = NULL;

	if (test_mode) {
		/* override -afn flags if -t was specified */
		async = 0;
		fork_mode = 0;
		lockmode = F_GETLK;
	}

	return childargv;
}

/*
 *
 * main
 *
 */

int main(int argc, char *argv[]) {
	char **childargv = NULL;
	int child_return = 99;
	int lockfd;
	struct flock flock_struct;
	int lockfd_flags;
	pid_t async_fork = 0;

	cmdname = argv[0];

	if (argc < 2) {
		usage();
	}
	
	/*
	 *
	 * Command line parsing
	 *
	 */

	childargv = cmd_line(argc, argv);

	

	/*
	 *
	 * Open the file
	 *
	 */

	lockfd = open(lockfile, openmode, 0600);

	if (lockfd == -1) {
		error("Can't open lockfile \"%s\": %s\n",
			lockfile, strerror(errno));
	}

	/*
	 *
	 * Set the Close on Exec flag
	 *
	 */
	

	if (fork_mode) {	
		lockfd_flags = fcntl(lockfd, F_GETFD);
		lockfd_flags |= FD_CLOEXEC;
		fcntl(lockfd, F_SETFD, lockfd_flags);
	}

	/*
	 *
	 * get the exclusive lock
	 *
	 */

	flock_struct.l_type = F_WRLCK;
	flock_struct.l_whence = SEEK_SET;
	flock_struct.l_start = 0;
	flock_struct.l_len = 0;

	if (fcntl(lockfd, lockmode, &flock_struct) == -1) {
		if ((lockmode == F_SETLK) && (errno == EAGAIN)) {
			/* nonblocking failed */
			exit(FAIL_EXITCODE_NB);
		}
		if (lockmode == F_GETLK) {
			/* test mode */
			error("Can't get lock information on \"%s\": %s\n",
				lockfile, strerror(errno));
		}
		error("Can't get exclusive lock on \"%s\": %s\n",
			lockfile, strerror(errno));
	}

	/*
	 *
	 * fork() for async mode
	 *
	 */

	if (async) {
		async_fork = fork();
		if (async_fork == -1) {
			/* fail */
			error("fork() failed: %s\n", strerror(errno));
		}
		if (async_fork > 0 ) {
			/* daddy dies */
			exit(0);
		}
	}

	/*
	 *
	 * do the work
	 *
	 */

	if (! test_mode) {
		child_return = mk_child(fork_mode, childargv);
	} else {
		/* test mode */
		if (flock_struct.l_type == F_UNLCK) {
			child_return = 0;
		} else {
			printf("%ld\n", (long)flock_struct.l_pid);
			child_return = FAIL_EXITCODE_NB;
		}
	}

	if (! test_mode) {
		/*
		 *
		 * unlock
		 *
		 */

		flock_struct.l_type = F_WRLCK;
		flock_struct.l_whence = SEEK_SET;
		flock_struct.l_start = 0;
		flock_struct.l_len = 0;
		fcntl(lockfd, F_SETLKW, &flock_struct); /* errors?? -> bad luck */
	}

	/*
	 *
	 * close the lockfile
	 *
	 */

	close(lockfd);	

	exit(child_return);
}

